#!/usr/bin/env python3
import os
import sys
import time
from argparse import ArgumentParser
from hashlib import sha1
from pathlib import Path
from subprocess import PIPE, DEVNULL, list2cmdline, run, call, check_call
from tempfile import TemporaryDirectory
from typing import Union, Optional

BASE_PATH = Path(__file__).parent
SHA1SUMS_PATH = BASE_PATH / "SHA1SUMS"
FW_PATH = BASE_PATH / "estkme-red.fwupd"
VERSION_PATH = BASE_PATH / "VERSION"
PRODUCT_NAME = "ESTKme-RED"

Upgrader = Union["DarwinUpgrader", "LinuxUpgrader"]


class Version(tuple[int, ...]):
    @classmethod
    def parse(cls, version: str):
        return cls(map(int, version.split(".")))

    def __str__(self):
        return ".".join(map(str, self))


def get_version(program: Path) -> Optional[Version]:
    process = run([program, "ver"], stdout=PIPE, stderr=DEVNULL, universal_newlines=True)
    stdout = process.stdout.strip()
    if not stdout:
        return None
    return Version.parse(stdout)


# noinspection SpellCheckingInspection
class DarwinUpgrader:
    program: Path
    volume: Path

    def __init__(self, volume: str):
        self.program = BASE_PATH / "bin" / f"redctl.darwin"
        self.program.chmod(0o755)
        self.volume = Path("/Volumes") / volume

    def is_mounted(self) -> bool:
        return self.volume.exists()

    def upgrade(self):
        if not self.is_mounted():
            call([str(self.program), "dfu"])
        while not self.is_mounted():
            time.sleep(1)
        mount_point = str(self.volume)
        commands = "\n".join(map(list2cmdline, (
            ("set", "-xeuo", "pipefail"),
            ("cp", "-v", FW_PATH.name, mount_point),
            ("diskutil", "unmountDisk", mount_point),
        )))
        check_call(["bash", "-c", commands], cwd=str(FW_PATH.parent))


# noinspection SpellCheckingInspection
class LinuxUpgrader:
    program: Path
    disk: Path

    def __init__(self, label: str, machine: str):
        self.program = BASE_PATH / "bin" / f"redctl.linux-{machine}"
        self.program.chmod(0o755)
        self.disk = Path("/dev/disk/by-label") / label

    def is_mounted(self) -> bool:
        return self.disk.exists()

    def upgrade(self):
        if not self.is_mounted():
            call([str(self.program), "dfu"], stderr=DEVNULL)
        while not self.is_mounted():
            time.sleep(1)
        with TemporaryDirectory() as mount_point:
            disk = str(self.disk.resolve())
            commands = "\n".join(map(list2cmdline, (
                ("set", "-xeuo", "pipefail"),
                ("mount", "-v", disk, mount_point),
                ("cp", "-v", FW_PATH.name, mount_point),
                ("umount", "-v", disk),
            )))
            check_call(["sudo", "bash", "-c", commands], cwd=str(FW_PATH.parent))


def initialize_upgrader() -> Upgrader:
    uname = os.uname()
    if uname.sysname == "Darwin":
        return DarwinUpgrader(PRODUCT_NAME)
    elif uname.sysname == "Linux":
        return LinuxUpgrader(PRODUCT_NAME, uname.machine)
    else:
        raise Exception(f"Unsupported system: {uname.sysname}")


def integrity_check():
    if not SHA1SUMS_PATH.exists():
        print("Now the integrity verification been BROKEN, upgrading is RISKY.")
        confirm_text = input("Continue upgrading? (Y/N) ").upper()
        if confirm_text not in {"Y", "YES"}:
            raise Exception("Upgrade Cancelled")
        return
    with SHA1SUMS_PATH.open("r") as fp:
        rows = [
            (bytes.fromhex(expected_hash), BASE_PATH / filename)
            for expected_hash, filename in map(str.split, fp)
        ]
    tested = True
    for expected_hash, file in rows:
        rel_path = file.relative_to(BASE_PATH)
        if not file.exists():
            print(f"{rel_path} is missing", file=sys.stderr)
            tested = False
        elif expected_hash != sha1(file.read_bytes()).digest():
            print(f"{rel_path} is corrupted", file=sys.stderr)
            tested = False
    if not tested:
        raise Exception("Integrity Check: FAILED")


def read_version(upgrader: Upgrader):
    version = None
    while not version:
        version = get_version(upgrader.program)
        time.sleep(1)
    return version


def upgrade(force: bool):
    upgrader: Upgrader = initialize_upgrader()

    old_version = get_version(upgrader.program)
    new_version = Version.parse(VERSION_PATH.read_text())

    if not force and not upgrader.is_mounted():
        if old_version is None:
            raise Exception(f"Please connect {PRODUCT_NAME} to computer")
        if old_version == (0, 0, 0):
            raise Exception("\033[0;31mREQUIRE USE short-circuit update firmware\033[0m")
        if old_version >= new_version:
            raise Exception(f"{PRODUCT_NAME} is already up to date ({new_version})")

    if old_version:
        print("Old firmware version:", old_version)
    print("New firmware version:", new_version)

    print(f"Enter {PRODUCT_NAME} firmware update mode")
    upgrader.upgrade()
    print(f"\033[0;31mPlease re-plug {PRODUCT_NAME} to continue\033[0m")

    cur_version = read_version(upgrader)

    if new_version == cur_version:
        print("\033[0;32mFirmware update successfully\033[0m")
    else:
        print("\033[0;31mFirmware update failed\033[0m")


def main():
    parser = ArgumentParser()
    parser.add_argument("--force", action="store_true")
    args = parser.parse_args()
    upgrade(force=args.force)


if __name__ == "__main__":
    try:
        integrity_check()
        main()
    except KeyboardInterrupt:
        pass
    except Exception as e:
        print(e)
